﻿/*======================================================================================
    Copyright 2021 by TheDummyProgrammer (https://www.thedummyprogrammer.com)

    This file is part of The Dummy Programmer Robot.

    The Dummy Programmer Robot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The Dummy Programmer Robot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Dummy Programmer Robot.  If not, see <http://www.gnu.org/licenses/>.
======================================================================================*/

using Microsoft.VisualBasic.FileIO;
using System;
using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using TDP.Robot.Core;
using TDP.Robot.Core.DynamicData;
using TDP.Robot.Core.Logging;

namespace TDP.Robot.Plugins.Core.ReadTextFileTask
{
    [Serializable]
    public class ReadTextFileTask : ITask
    {
        public IFolder ParentFolder { get; set; }
        public int ID { get; set; }
        public IPluginInstanceConfig Config { get; set; }

        public List<PluginInstanceConnection> Connections { get; } = new List<PluginInstanceConnection>();


        public void Init()
        {

        }

        public void Destroy()
        {
            
        }

        private string[] BuildDelimitersArray(ReadTextFileTaskConfig config)
        {
            List<string> Delimiters = new List<string>();

            if (config.DelimiterTab)
                Delimiters.Add(@"\t");
            
            if (config.DelimiterSemicolon)
                Delimiters.Add(@";");

            if (config.DelimiterComma)
                Delimiters.Add(@",");

            if (config.DelimiterSpace)
                Delimiters.Add(@" ");

            if (config.DelimiterOther)
                Delimiters.Add(config.DelimiterOtherChar);

            return Delimiters.ToArray();
        }

        private int[] BuildColumnsWidthArray(ReadTextFileTaskConfig config)
        {
            int[] ColWidths = new int[config.ColumnsWidth.Count];

            for (int i = 0; i < config.ColumnsWidth.Count - 1; i++)
            {
                ColWidths[i] = int.Parse(config.ColumnsWidth[i].ColumnWidth);
            }

            return ColWidths;
        }

        private string[] ReadRowFromParser(TextFieldParser parser, TextFileReadSplitColumnsType splitType)
        {
            if (splitType == TextFileReadSplitColumnsType.None)
                return new string[] { parser.ReadLine() };
            else
                return parser.ReadFields();
        }

        private void AdjustDataTableColumnsCount(DataTable dt, int columnsCount)
        {
            if (dt.Columns.Count < columnsCount)
            {
                int InitialColCount = dt.Columns.Count;
                for (int i = 1; i <= columnsCount - InitialColCount; i++)
                {
                    dt.Columns.Add("Column" + i.ToString(), typeof(string));
                }
            }
        }

        public ExecResult Run(DynamicDataChain dataChain, DynamicDataSet lastDynamicDataSet, IPluginInstanceLogger instanceLogger)
        {
            ExecResult Result;
            DateTime StartDateTime = DateTime.Now;

            DataTable DefaultRecordset = new DataTable();

            int ActualIterations = 0;

            try
            {
                if (!Config.DoNotLog)
                    instanceLogger.TaskStarted(this);

                ReadTextFileTaskConfig TConfig = (ReadTextFileTaskConfig)Config;
                int IterationsCount = DynamicDataParser.GetIterationCount(TConfig, dataChain, lastDynamicDataSet);

                if (IterationsCount > 0)
                {
                    // For this kind of objects consider only one iteration
                    ReadTextFileTaskConfig ConfigCopy = (ReadTextFileTaskConfig)CoreHelpers.CloneObjects(Config);
                    DynamicDataParser.Parse(ConfigCopy, dataChain, 0);

                    using (TextFieldParser FileParser = new TextFieldParser(ConfigCopy.FilePath))
                    {
                        if (ConfigCopy.SplitColumnsType == TextFileReadSplitColumnsType.UseDelimiters)
                        {
                            FileParser.TextFieldType = FieldType.Delimited;
                            FileParser.Delimiters = BuildDelimitersArray(ConfigCopy);
                            FileParser.HasFieldsEnclosedInQuotes = ConfigCopy.UseDoubleQuotes;
                        }
                        else if (ConfigCopy.SplitColumnsType == TextFileReadSplitColumnsType.UseFixedWidthColumns)
                        {
                            FileParser.TextFieldType = FieldType.FixedWidth;
                            FileParser.FieldWidths = BuildColumnsWidthArray(ConfigCopy);
                        }

                        int ReadFromRow = 0;
                        int ReadToRow = 0;

                        if (TConfig.ReadAllTheRowsOption)
                        {
                            string[] Row = null;
                            while (!FileParser.EndOfData)
                            {
                                try
                                {
                                    Row = ReadRowFromParser(FileParser, ConfigCopy.SplitColumnsType);
                                    AdjustDataTableColumnsCount(DefaultRecordset, Row.Length);
                                    DefaultRecordset.Rows.Add(Row);
                                }
                                catch (MalformedLineException ex)
                                {
                                    instanceLogger.Error(this, "Error parsing line", ex);
                                }
                            }
                        }
                        else if (TConfig.ReadLastRowOption)
                        {
                            string[] Row = null;
                            while (!FileParser.EndOfData)
                            {
                                try
                                {
                                    Row = ReadRowFromParser(FileParser, ConfigCopy.SplitColumnsType);
                                }
                                catch (MalformedLineException ex)
                                {
                                    instanceLogger.Error(this, "Error parsing line", ex);
                                }
                            }
                            if (Row != null)
                            {
                                AdjustDataTableColumnsCount(DefaultRecordset, Row.Length);
                                DefaultRecordset.Rows.Add(Row);
                            }
                        }
                        else if (TConfig.ReadRowNumberOption || (TConfig.ReadIntervalOption && TConfig.ReadInterval == TextFileReadIntervalType.ReadFromRowToRow))
                        {
                            if (TConfig.ReadRowNumberOption)
                            {
                                ReadFromRow = int.Parse(ConfigCopy.ReadRowNumber);
                                ReadToRow = ReadFromRow;
                            }
                            else
                            {
                                ReadFromRow = int.Parse(ConfigCopy.ReadFromRow);
                                ReadToRow = int.Parse(ConfigCopy.ReadToRow);
                            }

                            string[] Row = null;
                            int RowIndex = 1;
                            while (!FileParser.EndOfData)
                            {
                                try
                                {
                                    if (RowIndex > ReadToRow)
                                        break;

                                    Row = ReadRowFromParser(FileParser, ConfigCopy.SplitColumnsType);
                                    if (RowIndex >= ReadFromRow && RowIndex <= ReadToRow)
                                    {
                                        AdjustDataTableColumnsCount(DefaultRecordset, Row.Length);
                                        DefaultRecordset.Rows.Add(Row);
                                    }
                                }
                                catch (MalformedLineException ex)
                                {
                                    instanceLogger.Error(this, "Error parsing line", ex);
                                }

                                RowIndex++;
                            }
                        }
                        else
                        {
                            List<string[]> Rows = new List<string[]>();
                            string[] Row = null;
                            while (!FileParser.EndOfData)
                            {
                                try
                                {
                                    Row = ReadRowFromParser(FileParser, ConfigCopy.SplitColumnsType);
                                    AdjustDataTableColumnsCount(DefaultRecordset, Row.Length);
                                    Rows.Add(Row);
                                }
                                catch (MalformedLineException ex)
                                {
                                    instanceLogger.Error(this, "Error parsing line", ex);
                                }
                            }

                            if (Rows.Count > 0)
                            {
                                if (TConfig.ReadInterval == TextFileReadIntervalType.ReadFromRowToLastRow)
                                {
                                    // From row to end
                                    ReadFromRow = int.Parse(ConfigCopy.ReadFromRow);
                                    ReadFromRow--;
                                    if (ReadFromRow < 0) ReadFromRow = 0;
                                }
                                else
                                {
                                    // Read N Last rows
                                    ReadFromRow = (Rows.Count - 1) - int.Parse(ConfigCopy.ReadNumberOfRows);
                                }

                                if (ReadFromRow <= (Rows.Count - 1))
                                {
                                    for (int RowIndex = ReadFromRow; RowIndex < Rows.Count; RowIndex++)
                                    {
                                        DefaultRecordset.Rows.Add(Rows[RowIndex]);
                                    }
                                }
                            }
                        }
                    }
                }

                DynamicDataSet DDataSet = CommonDynamicData.BuildStandardDynamicDataSet(this, true, 0, StartDateTime, DateTime.Now, ActualIterations);

                DDataSet.Add(CommonDynamicData.DefaultRecordsetName, DefaultRecordset);
                Result = new ExecResult(true, DDataSet);

                if (!Config.DoNotLog)
                    instanceLogger.TaskCompleted(this);
            }
            catch (Exception ex)
            {
                if (!Config.DoNotLog)
                    instanceLogger.TaskError(this, ex);

                DynamicDataSet DDataSet = CommonDynamicData.BuildStandardDynamicDataSet(this, false, -1, StartDateTime, DateTime.Now, ActualIterations);
                DDataSet.Add(CommonDynamicData.DefaultRecordsetName, DefaultRecordset);
                Result = new ExecResult(false, DDataSet);
            }

            return Result;
        }
    }
}
